# _external_archiver_base.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



""":class:`_ExternalArchiverBase` class."""



__all__ = ["_ExternalArchiverBase"]



# {{{ INCLUDES

from abc import *
import os
import subprocess
import select

from ._archiver_base import *

# }}} INCLUDES



# {{{ CLASSES

class _ExternalArchiverBase(_ArchiverBase):
    """Base class for external archivers.

    See also: :class:`._ArchiverBase`."""

    #: Full path to the archiver program.  Type: ``str``.
    ARCHIVER_ = None



    # {{{ overrides of _ArchiverBase

    @abstractmethod
    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        # derived classes has to specify self.ARCHIVER_
        assert self.ARCHIVER_



    def executeArchiver_(self):
        """See: :meth:`._ArchiverBase.executeArchiver_()`.

        Method calls :meth:`arguments_()`, then it executes the :term:`archiver`.  Each archiver's message is shown
        via :meth:`showArchiverMessage_()` method.  In case the :term:`archiver` returns non-zero exit code, the
        :meth:`handleArchiverExitCode_()` method is called."""

        status = True

        sysEnvironment = os.environ.copy()
        arguments = self.arguments_(sysEnvironment)
        sysEnvironment.update(arguments[1])
        try:
            archiverProcess = subprocess.Popen([self.ARCHIVER_] + arguments[0], stdout = subprocess.PIPE,
                                               stderr = subprocess.PIPE, env = sysEnvironment,
                                               universal_newlines = True)
        except OSError as ex:
            self.componentUi_.showError(str.format(
                    "Attempt to execute external archiving program \"{}\" failed. The returned error was: \"{}\".",
                    self.ARCHIVER_, ex))
            return False

        # capture program's standard output and standard error and use IComponentUi interface to print captured
        # messages; note that the order of messages written to stdout vs. messages written to stderr may not be
        # preserved
        while True:
            readyStreams = select.select((archiverProcess.stdout, archiverProcess.stderr), (), ())[0]
            streamActive = False
            for readyStream in readyStreams:
                line = readyStream.readline()
                if line:
                    streamActive = True
                    self.showArchiverMessage_(line[:-1], readyStream is not archiverProcess.stdout)
            if archiverProcess.poll() is not None and not streamActive:
                break

        if archiverProcess.returncode:
            status = self.handleArchiverExitCode_(archiverProcess.returncode)

        return status

    # }}} overrides of _ArchiverBase



    # {{{ virtual members

    @abstractmethod
    def arguments_(self, sysEnvironment):
        """Assemble arguments and environment for the archiver program.

        Gets all necessary information, typically from the environment, configured options and from the archive
        specification stored in :attr:`._AbstractArchiver.archiveSpec_`, and sets the environment and assembles all
        options and arguments that should be passed to the :term:`archiver` program in order to create the desired
        :term:`backup`.

        :param sysEnvironment: Application's system environment as a dictionary of environment variables as keys and
           their values as values.
        :type sysEnvironment: ``dict<str: str>``

        :return: Tuple of archiver's arguments and an environment dictionary.  Arguments are stored in the list of
           strings and the environment is the dictionary where keys are env. variable names and values are their
           corresponding values.
        :rtype: ``tuple[list<str>, dict<str: str>]``"""



    def showArchiverMessage_(self, message, sentToStderr = False):
        """Shows archiver message as an error, warning, notification, etc.

        If the message was sent to stdout (``sentToStderr`` == ``False``) then it will be shown as
        *verbose message*.  Otherwise it will be shown as *warning message*\ .

        :param message: Message produced by an archiver.
        :type message: ``str``
        :param sentToStderr: ``True`` if ``message`` was sent to stderr by the archiver.
        :type sentToStderr: ``bool``"""

        if not sentToStderr:
            self.componentUi_.showVerbose(message)
        else:
            self.componentUi_.showWarning(message)

    # }}} virtual members

# }}} CLASSES
