# _incremental_archiver_mixin.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



""":class:`_IncrementalArchiverMixin` class."""



__all__ = ["_IncrementalArchiverMixin"]



# {{{ INCLUDES

from abc import *
import os.path
from datetime import date, datetime

from ..._configuration import *
from .. import *
from ._archive_spec import *
from ._abstract_incremental_archiver import *

# }}}



# {{{ CLASSES

class _IncrementalArchiverMixin(_AbstractIncrementalArchiver):
    """Defines functionality for incremental archivers.

    Abstract constructor of this class, should be called from derived constructors.

    .. note:: A class extended with this mixin has to derive from :class:`_AbstractArchiver`."""

    # storage variables for backup level restarting
    _STORAGE_LAST_RESTART = "archiving-last-restart"
    _STORAGE_LAST_FULL_RESTART = "archiving-last-full-restart"
    _STORAGE_BACKUP_SIZE = "archiving-size-level"
    _STORAGE_RESTART_COUNT = "archiving-restart-count"



    # {{{ _AbstractIncrementalArchiver implementation

    @abstractmethod
    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        self.__nextBackupLevel = None
        self.__restartReason = None
        self.__restartLevel = None



    @property
    def nextBackupLevel(self):
        "See: :attr:`._AbstractIncrementalArchiver.nextBackupLevel`."

        return self.__nextBackupLevel



    @property
    def restartReason(self):
        "See: :attr:`._AbstractIncrementalArchiver.restartReason`."

        return self.__restartReason



    @property
    def restartLevel(self):
        "See: :attr:`._AbstractIncrementalArchiver.restartLevel`."

        return self.__restartLevel



    def getMaxBackupLevel(self):
        "See: :meth:`._AbstractIncrementalArchiver.getMaxBackupLevel()`."

        currentBackupLevel = self.getBackupLevelForArchive(
            self.archiveSpec_[_ArchiveSpecOptions.NAME], self.archiveSpec_[Options.USER_CONFIG_DIR])
        return currentBackupLevel + 1 if currentBackupLevel is not None else 0



    @classmethod
    @abstractmethod
    def getBackupLevelForArchive(cls, archiveName, userConfigDir):
        "See: :meth:`._AbstractIncrementalArchiver.getBackupLevelForArchive()`."

        return None

    # }}} _AbstractIncrementalArchiver implementation



    # {{{ helpers

    def setNextBackupLevel_(self):
        """Sets the :attr:`nextBackupLevel` and :attr:`restartReason` properties.

        .. attention:: This method has to be called from :meth:`._AbstractArchiver.attachArchiveSpec()` after \
           :meth:`setRestartLevel_()`."""

        level = self.getMaxBackupLevel()
        reason = BackupLevelRestartReasons.NoRestart

        if self.archiveSpec_[Options.LEVEL] is not None:
            if self.archiveSpec_[Options.LEVEL] > level:
                self.componentUi_.showWarning(str.format(
                        "Backup level {} value too high. Using level {} instead, which is next in a row.",
                        self.archiveSpec_[Options.LEVEL], level))
            else:
                level = self.archiveSpec_[Options.LEVEL]

        elif self.archiveSpec_[Options.RESTARTING]:
            level, reason = self.__restartBackupLevel(level)

        self.__nextBackupLevel = level
        self.__restartReason = reason



    def clearNextBackupLevel_(self):
        """Clears the :attr:`nextBackupLevel` and :attr:`restartReason` properties.

        .. note:: This method should be called from :meth:`._AbstractArchiver.detachArchiveSpec()`."""

        self.__nextBackupLevel = None
        self.__restartReason = None



    def getLevelSuffix_(self, level):
        """Returns file name suffix according to the backup level.

        .. note:: \ :attr:`._AbstractArchiver.archiveSpec_` has to be initialized.

        :param level: Backup level for which the suffix shall be returned.
        :type level: ``int``

        :return: File name suffix for backup level in the form '.<level>'.
        :rtype: ``str``"""

        levelSuffix = ""
        if self.archiveSpec_[Options.INCREMENTAL] and level > 0:
            levelSuffix = "." + str(level)
        return levelSuffix



    @staticmethod
    def getLevelFromFileName_(fileName):
        """Extracts backup level number from the file name.

        :param fileName: Name of the file used to get the backup level from.  It should be in the form:
          '<archive_name>[.<level>].<suffix>'.
        :type fileName: ``str``

        :return: The :term:`backup level` retrieved from the file name.
        :rtype: ``int``"""

        try:
            levelToken = os.path.splitext(os.path.splitext(fileName)[0])[1]
            level = int(levelToken[1:])
        except ValueError:
            level = 0

        return level



    # {{{ backup level restarting

    def setRestartLevel_(self):
        """Sets the :attr:`restartLevel` property.

        .. warning:: This method has to be called from :meth:`._AbstractArchiver.attachArchiveSpec()`."""

        maxRestartLevelSize = self.archiveSpec_[Options.MAX_RESTART_LEVEL_SIZE]

        if not maxRestartLevelSize:
            self.__restartLevel = 1
            return

        try:
            level0Size = self.__getArchiveSize()
            restartLevel = 1
            while (self.__getArchiveSize(restartLevel) / level0Size) * 100 > maxRestartLevelSize and \
                  restartLevel < self.getMaxBackupLevel() - 1:
                restartLevel += 1
        except KeyError:
            restartLevel = 1
            self.componentUi_.showWarning(str.format(
                    "Unable to get the size of backup files in order to determine the restart backup level. Falling " +
                    "back to backup level {}.", restartLevel))

        self.__restartLevel = restartLevel



    def clearRestartLevel_(self):
        """Clears the :attr:`restartLevel` property.

        .. note:: This method should be called from :meth:`._AbstractArchiver.detachArchiveSpec()`."""

        self.__restartLevel = None



    def reportBackupLevelRestart_(self):
        "Shows a user message if :term:`backup level` was restarted."

        if self.restartReason is BackupLevelRestartReasons.RestartCountLimitReached:
            self.componentUi_.showWarning("Maximal backup level restart count reached. Restarting to level 0.")
        elif self.restartReason is BackupLevelRestartReasons.LastFullRestartAgeLimitReached:
            self.componentUi_.showWarning("Maximal backup level full restart age reached. Restarting to level 0.")
        elif self.restartReason is BackupLevelRestartReasons.BackupLevelLimitReached:
            self.componentUi_.showWarning(str.format("Maximal backup level reached. Restarting to level {}.",
                                                     self.nextBackupLevel))
        elif self.restartReason is BackupLevelRestartReasons.LastRestartAgeLimitReached:
            self.componentUi_.showWarning(str.format(
                "Maximal backup level restart age reached. Restarting to level {}.", self.nextBackupLevel))



    def saveRestartingInfo_(self):
        """Saves all the information required for backup level restarting to the persistent storage.

        This method has to be called after the :term:`backup` is successfully created."""

        if not self.archiveSpec_[Options.RESTARTING]:
            return

        # increase and save restart count
        if self.restartReason is not BackupLevelRestartReasons.NoRestart:
            restartCount = int(self.storagePortion_.getValue(self._STORAGE_RESTART_COUNT))
            self.storagePortion_.saveValue(self._STORAGE_RESTART_COUNT, str(restartCount + 1))

        # save last restart date
        if self.nextBackupLevel < 2 or self.restartReason is not BackupLevelRestartReasons.NoRestart:
            self.storagePortion_.saveValue(self._STORAGE_LAST_RESTART, date.today())
            if self.nextBackupLevel == 0:
                self.storagePortion_.saveValue(self._STORAGE_LAST_FULL_RESTART, date.today())
                self.storagePortion_.saveValue(self._STORAGE_RESTART_COUNT, "0")

        # save the size of the created backup
        archiveSize = os.path.getsize(self.archiveFileName_)
        self.storagePortion_.saveValue(self._STORAGE_BACKUP_SIZE + str(self.nextBackupLevel), archiveSize)

        # >clear stored backup sizes for higher levels
        levelIdx = self.nextBackupLevel + 1
        while self.storagePortion_.tryRemoveVariable(self._STORAGE_BACKUP_SIZE + str(levelIdx)):
            levelIdx += 1



    @staticmethod
    def _getRestartDate(storageVariable, storagePortion):
        "Reads a date of last backup level restart from file and returns it."

        return datetime.strptime(storagePortion.getValue(storageVariable), "%Y-%m-%d").date()



    def __restartBackupLevel(self, nextLevel):
        """Restart the passed ``nextLevel`` backup level if needed.

        :param nextLevel: Next backup level if it would not be restarted.
        :type nextLevel: ``int``

        :return: Backup level after restart and the reason for the restart.
        :rtype: ``tuple[int, BackupLevelRestartReasons]``"""

        # full restart handling
        if nextLevel < 1:
            return nextLevel, BackupLevelRestartReasons.NoRestart

        if self.archiveSpec_[Options.FULL_RESTART_AFTER_COUNT]:
            if int(self.storagePortion_.getValue(self._STORAGE_RESTART_COUNT)) >= \
               self.archiveSpec_[Options.FULL_RESTART_AFTER_COUNT]:
                return 0, BackupLevelRestartReasons.RestartCountLimitReached

        if self.archiveSpec_[Options.FULL_RESTART_AFTER_AGE]:
            if not self.storagePortion_.hasVariable(self._STORAGE_LAST_FULL_RESTART):
                self.componentUi_.showError("Unable to read the last full restart date.")
                return nextLevel, BackupLevelRestartReasons.NoRestart

            if (date.today() - self._getRestartDate(self._STORAGE_LAST_FULL_RESTART, self.storagePortion_)).days > \
                    self.archiveSpec_[Options.FULL_RESTART_AFTER_AGE]:
                return 0, BackupLevelRestartReasons.LastFullRestartAgeLimitReached

        # standard restart handling
        if nextLevel < 2:
            return nextLevel, BackupLevelRestartReasons.NoRestart

        if nextLevel > self.archiveSpec_[Options.RESTART_AFTER_LEVEL]:
            return self.restartLevel, BackupLevelRestartReasons.BackupLevelLimitReached

        if self.archiveSpec_[Options.RESTART_AFTER_AGE]:
            if not self.storagePortion_.hasVariable(self._STORAGE_LAST_RESTART):
                self.componentUi_.showError("Unable to read the last restart date.")
                return nextLevel, BackupLevelRestartReasons.NoRestart

            if (date.today() - self._getRestartDate(self._STORAGE_LAST_RESTART, self.storagePortion_)).days > \
                    self.archiveSpec_[Options.RESTART_AFTER_AGE]:
                return self.restartLevel, BackupLevelRestartReasons.LastRestartAgeLimitReached

        return nextLevel, BackupLevelRestartReasons.NoRestart



    def __getArchiveSize(self, level = 0):
        return int(self.storagePortion_.getValue(self._STORAGE_BACKUP_SIZE + str(level)))

    # }}} backup level restarting

    # }}} helpers

# }}} CLASSES
