# _internal_archiver_base.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



""":class:`_InternalArchiverBase` class."""



__all__ = ["_InternalArchiverBase"]



# {{{ INCLUDES

from abc import *
import os
import stat
import itertools

from ..._configuration import *
from ._archiver_base import *
from ._archive_spec import *

# }}} INCLUDES



# {{{ CLASSES

class _InternalArchiverBase(_ArchiverBase):
    """Base class for internal archivers.

    See also: :class:`._ArchiverBase`."""

    # {{{ overrides of _ArchiverBase

    def archiveFileName_(self):
        "See: :attr:`._AbstractArchiver.archiveFileName_`."

        return self.__getArchiveFileName()



    def executeArchiver_(self):
        """See: :meth:`._ArchiverBase.executeArchiver_()`.

        Adds filesystem objects to the backup.  It uses :meth:`addRegularFile_()` and :meth:`addSpecialFile_()` methods
        for that.  In case of an error the :meth:`._ArchiverBase.handleArchiverExitCode_()` is called."""

        status = True
        
        try:
            cwdSave = os.getcwd()
            os.chdir(self.archiveSpec_[_ArchiveSpecOptions.PATH])
            for name in self.__getFileNames(self.archiveSpec_[_ArchiveSpecOptions.INCLUDE_FILES],
                                            self.archiveSpec_[_ArchiveSpecOptions.EXCLUDE_FILES]):
                try:
                    statInfo = os.lstat(name)
                except OSError as ex:
                    self.componentUi_.showError(str.format("Cannot access \"{}\". {}.", name, ex.strerror))
                    status = False
                    continue
                if stat.S_ISREG(statInfo[stat.ST_MODE]):
                    try:
                        self.addRegularFile_(name, statInfo)
                    except IOError as ex:
                        self.componentUi_.showError(str.format("Cannot open file \"{}\". {}.", name, ex.strerror))
                        status = False
                else:
                    self.addSpecialFile_(name, statInfo)

            os.chdir(cwdSave)
        except OSError as ex:
            self.componentUi_.showError(str.format("System error occurred: {}.", ex.strerror))
            status = False

        if not status:
            self.handleArchiverExitCode_(1)

        return status

    # }}} overrides of _ArchiverBase



    # }}} virtual members

    @abstractmethod
    def addRegularFile_(self, name, statInfo):
        """Adds a regular file to the :term:`backup`.

        Method is called for each regular file that shall be added to the backup.

        .. note:: Implementation should call :meth:`showFileNameVerbose_()` just before the file is added to the backup,
           when no error is expected anymore.

        See also: :meth:`addSpecialFile_()`.

        :param name: Path to a regular file that shall be added to the backup.  It should be relative to archive's
           base directory (:attr:`._ArchiveSpecOptions.PATH`).
        :type name: ``str``
        :param statInfo: Stat information of the file ``name`` as returned by :meth:`os.lstat()`.
        :type statInfo: ``stat_result``

        :raise IOError: If there was an error while opening the file."""



    @abstractmethod
    def addSpecialFile_(self, name, statInfo):
        """Adds a special file to the :term:`backup`.

        Method is called for each special file, such as directory, named pipe, device file, link et al. that shall be
        added to the backup.

        .. note:: Implementation should call :meth:`showFileNameVerbose_()` just before the file is added to the backup,
           when no error is expected anymore.

        See also: :meth:`addRegularFile_()`.

        :param name: Path to a special file that shall be added to the backup.  It should be relative to archive's
           base directory (:attr:`._ArchiveSpecOptions.PATH`).
        :type name: ``str``
        :param statInfo: Stat information of the file ``name`` as returned by :meth:`os.lstat()`.
        :type statInfo: ``stat_result``"""

    # }}} virtual members



    # {{{ helpers

    def showFileNameVerbose_(self, name):
        if self.archiveSpec_[Options.VERBOSE]:
            self.componentUi_.showVerbose(name)



    def __getArchiveFileName(self):
        """Gets path to the file that shall be created by the :term:`archiver` program.

        :return: Path to the backup file.
        :rtype: ``str``"""

        destDir = self.archiveSpec_[Options.DEST_DIR]
        fileName = self.archiveSpec_[_ArchiveSpecOptions.NAME] + self.suffix_

        return os.path.join(destDir, fileName)



    def __getFileNames(self, includeFiles, excludeFiles):

        def getFileNamesUnfiltered(includeFiles):
            for includeName in includeFiles:

                if os.path.isdir(includeName) and not os.path.islink(includeName):

                    # variables to track if includeName and directories list returned by os.walk() were visited in the
                    # os.walk() for loop; if not, assume permission denied
                    includeNameVisited = False
                    listedDirs = set()
                    visitedDirs = set()

                    for root, dirs, files in os.walk(includeName):
                        includeNameVisited = True
                        listedDirs.update({os.path.join(root, d)
                                           for d in dirs
                                           if not os.path.islink(os.path.join(root, d))})
                        visitedDirs.add(root)

                        yield root

                        # yield files and links to directories within root (we yield links to directories as they
                        # were files because we do not follow them)
                        for name in itertools.chain(files, (d for d in dirs if os.path.islink(os.path.join(root, d)))):
                            yield os.path.join(root, name)

                    if not includeNameVisited:
                        showDirOpenError(includeName)
                    for notVisitedDir in listedDirs - visitedDirs:
                        showDirOpenError(notVisitedDir)

                else:
                    yield includeName



        def showDirOpenError(dirName):
            if dirName not in excludeFiles:
                self.componentUi_.showError(str.format("Cannot open directory \"{}\". Permission denied.", dirName))



        def parentDirs(path, _result = None):
            """Returns a set of parent directories of ``path``.

            For example for "dir/foo/bar" it returns {"dir/foo", "dir"}."""

            if _result is None:
                _result = set()

            dirname = os.path.dirname(path)
            if dirname:
                _result.add(dirname)
                parentDirs(dirname, _result)
            return _result



        def filterFileNames(names, excludeFiles):
            return (name for name in names if (name not in excludeFiles and parentDirs(name).isdisjoint(excludeFiles)))



        return filterFileNames(getFileNamesUnfiltered(includeFiles), excludeFiles)

    # }}} helpers

# }}} CLASSES
