# _internal_archivers.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



"""Concrete classes for internal archivers."""



__all__ = ["_InternalTarArchiver", "_InternalTarCompressedArchiver", "_InternalTarGzArchiver",
           "_InternalTarBz2Archiver"]



# {{{ INCLUDES

from abc import *
import os
import stat
import tarfile

from ..._configuration import *
from .. import *
from ._internal_archiver_base import *

# }}} INCLUDES



# {{{ CLASSES

class _InternalTarArchiver(_InternalArchiverBase):
    "Represents internal tar archiver."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarInternal

    #: See: :attr:`._AbstractArchiver.SUFFIX_`.
    SUFFIX_ = ".tar"

    #: See: :attr:`._ArchiverBase.UNSUPPORTED_OPTIONS_`.
    UNSUPPORTED_OPTIONS_ = frozenset({Options.INCREMENTAL,
                                      Options.COMPRESSION_LEVEL})



    # {{{ overrides of _ArchiverBase

    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        self.__tarFile = None



    def preExecution_(self):
        "See: :meth:`._ArchiverBase.preExecution_()`."

        self.__tarFile = tarfile.open(self.archiveFileName_(), "w:" + self.tarFileOpenModeExtension_,
                                      format = tarfile.PAX_FORMAT)
        super().preExecution_()



    def postExecution_(self, success):
        "See: :meth:`._ArchiverBase.postExecution_()`."

        self.__tarFile.close()
        super().postExecution_(success)

    # }}} overrides of _ArchiverBase



    # {{{ overrides of _InternalArchiverBase

    def addRegularFile_(self, name, statInfo):
        "See: :meth:`._InternalArchiverBase.addRegularFile_()`."

        tarInfo = self.__tarFile.gettarinfo(name)

        with open(name, "rb") as fileObj:
            self.showFileNameVerbose_(name)
            self.__tarFile.addfile(tarInfo, fileObj)

            statInfoAfter = os.lstat(name)
            if statInfo[stat.ST_SIZE] != statInfoAfter[stat.ST_SIZE] or \
                statInfo[stat.ST_MTIME] != statInfoAfter[stat.ST_MTIME] or \
                statInfo[stat.ST_MODE] != statInfoAfter[stat.ST_MODE] or \
                statInfo[stat.ST_UID] != statInfoAfter[stat.ST_UID] or \
                statInfo[stat.ST_GID] != statInfoAfter[stat.ST_GID]:

                self.componentUi_.showWarning(str.format("File changed as we read it: \"{}\".", name))



    def addSpecialFile_(self, name, statInfo):
        "See: :meth:`._InternalArchiverBase.addSpecialFile_()`."

        tarInfo = self.__tarFile.gettarinfo(name)
        if tarInfo is not None:
            self.showFileNameVerbose_(name)
            self.__tarFile.addfile(tarInfo)
        else:
            if stat.S_ISSOCK(statInfo[stat.ST_MODE]):
                self.componentUi_.showWarning(str.format("Socket ignored: \"{}\".", name))
            else:
                self.componentUi_.showWarning(str.format("Unknown file type ignored: \"{}\".", name))

    # }}} overrides of _InternalArchiverBase



    # {{{ virtual members

    @property
    def tarFileOpenModeExtension_(self):
        """Gets the extension for tar file open mode.

        See also: :meth:`tarfile.open()`.

        :return: Extension for file open mode.
        :rtype: ``str``"""

        return ""

    # }}} virtual members



class _InternalTarCompressedArchiver(_InternalTarArchiver):
    "Represents internal tar archiver with a compression."

    #: Suffix specific for the compressed tar backup file.
    COMPRESS_SUFFIX_ = None

    #: Tar file open mode extension for creating a backup of a specific compression type.
    #: See also: :meth:`tarfile.open()`.
    COMPRESS_OPTION_ = None



    # {{{ overrides of _InternalTarArchiver

    @abstractmethod
    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

        # derived classes has to specify fields below
        assert self.COMPRESS_SUFFIX_ and self.COMPRESS_OPTION_



    @property
    def tarFileOpenModeExtension_(self):
        "See: :meth:`_InternalTarArchiver()`."

        return super().tarFileOpenModeExtension_ + self.COMPRESS_OPTION_

    # }}} overrides of _InternalTarArchiver



    # {{{ overrides of _ArchiverBase

    @property
    def suffix_(self):
        "See: :obj:`._ExternalArchiverBase.suffix_`."

        return super().suffix_ + self.COMPRESS_SUFFIX_

    # }}} overrides of _ArchiverBase



class _InternalTarGzArchiver(_InternalTarCompressedArchiver):
    "Represents internal tar archiver with gzip compression."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarGzInternal

    #: See: :attr:`_InternalTarCompressedArchiver.COMPRESS_SUFFIX_`.
    COMPRESS_SUFFIX_ = ".gz"

    #: See: :attr:`_InternalTarCompressedArchiver.COMPRESS_OPTION_`.
    COMPRESS_OPTION_ = "gz"



    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)



class _InternalTarBz2Archiver(_InternalTarCompressedArchiver):
    "Represents internal tar archiver with gzip compression."

    #: See: :attr:`._AbstractArchiver.ARCHIVER_TYPE`.
    ARCHIVER_TYPE = ArchiverTypes.TarBz2Internal

    #: See: :attr:`_InternalTarCompressedArchiver.COMPRESS_SUFFIX_`.
    COMPRESS_SUFFIX_ = ".bz2"

    #: See: :attr:`_InternalTarCompressedArchiver.COMPRESS_OPTION_`.
    COMPRESS_OPTION_ = "bz2"


    
    def __init__(self, interfaceAccessor):
        super().__init__(interfaceAccessor)

# }}} CLASSES
