# archiving_component.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2012 Róbert Čerňanský



""":class:`ArchivingComponent` class."""



__all__ = ["ArchivingComponent"]



# {{{ INCLUDES

import itertools

from ..._utils import *
from ..._mainf import *
from ..._configuration import *
from .. import *
from ._abstract_incremental_archiver import *
from ._archive_spec import *
from ._external_archivers import *
from ._internal_archivers import *
from ._incremental_archiver_mixin import *

# }}} INCLUDES



# {{{ CLASSES

class ArchivingComponent(IComponent, IArchiving):
    """Provides an interface for working with archives.

    During construction it registers itself as :class:`.IArchiving` component interface to ``interfaceAccessor``.

    Several methods of this class requires an :term:`archive specification file` as the input parameter (usually
    named ``specFile``).  This file should contain all information required to create the :term:`backup`.  Its format is
    defined by the standard :mod:`configparser` module.  It has to contain section ``[Content]`` and may contain
    section ``[Archive]``.  The ``[Content]`` section requires following options to be present: ``path``,
    ``include-files`` and ``exclude-files``.  Optionally, ``name`` can be present.  Options in the archive
    specification file has higher priority than those in the configuration."""

    # container of known archiver types; it should be sorted from most abstract to most concrete class
    __ARCHIVERS = (_ExternalTarArchiver,
                   _ExternalTarGzArchiver,
                   _ExternalTarBz2Archiver,
                   _ExternalTarLzmaArchiver,
                   _ExternalTarXzArchiver,
                   _InternalTarArchiver,
                   _InternalTarGzArchiver,
                   _InternalTarBz2Archiver)



    def __init__(self, interfaceAccessor):
        self.__interfaceAccessor = interfaceAccessor

        # stores already created _ArchiveSpec instances; key is path to the archive specification file and the value
        # is the instance
        self.__archiveSpecs = {}

        self.__appConfig = interfaceAccessor.getComponentInterface(IAppConfig)
        self.__storage = interfaceAccessor.getComponentInterface(IStorage)

        interfaceAccessor.registerComponentInterface(IArchiving, self)



    def run(self):
        "See: :meth:`.IComponent.run()`."

        return True



    # {{{ IArchiving implementation

    def makeBackup(self, specFile):
        """Creates the :term:`backup` based on ``specFile``.

        The result can be a file with a full backup or an incremental backup of some particular level.  This depends on
        the :term:`archive specification file` (``specFile``), the configuration (:class:`.IAppConfig`), previous
        operations with the ``specFile`` and the time.  Some of the properties of :class:`.ArchiveInfo` returned by the
        method :meth:`getArchiveInfo()` can be used to determine what the result will be.  The path and name of the
        created file will be assembled as follows:
        “<Options.DEST_DIR>/<archive_name>[.<backup_level>].<archiver_specific_extension>”.

        Method uses :class:`.IComponentUi` interface to report errors, warnings et al. to the user.

        .. warning:: This method utilizes the :term:`user configuration directory` so the option \
           :attr:`.Options.USER_CONFIG_DIR` has to point to an *existing* directory.

        See also: :meth:`.IArchiving.makeBackup()`."""

        archiveSpec = self.__getOrTryCreateArchiveSpec(specFile)
        if not archiveSpec:
            return

        archiver = self.__createArchiver(archiveSpec[Options.ARCHIVER])
        archiver.attachArchiveSpec(archiveSpec)
        archiver.createBackup()
        archiver.detachArchiveSpec()



    def filterValidSpecFiles(self, specFiles):
        "See: :meth:`.IArchiving.filterValidSpecFiles()`."

        def getOrTryCreateArchiveSpecSilently(specFile):
            try:
                return self.__getOrCreateArchiveSpec(specFile, True)
            except (IOError, LookupError, SyntaxError, KeyError, ValueError):
                return None



        archiveSpecs = (getOrTryCreateArchiveSpecSilently(archiveSpec)
                        for archiveSpec in specFiles)
        return (archiveSpec[_ArchiveSpecOptions.NAME]
                for archiveSpec in archiveSpecs
                if archiveSpec is not None)



    def getArchiveInfo(self, specFile):
        """See: :meth:`.IArchiving.getArchiveInfo()`

        .. warning:: This method utilizes the :term:`user configuration directory` so the option \
           :attr:`.Options.USER_CONFIG_DIR` has to point to an *existing* directory."""

        archiveSpec = self.__getOrTryCreateArchiveSpec(specFile)
        if not archiveSpec:
            return None

        storagePortion = self.__storage.createStoragePortion(realm = archiveSpec[_ArchiveSpecOptions.NAME])
        archiver = self.__createArchiver(archiveSpec[Options.ARCHIVER])
        archiver.attachArchiveSpec(archiveSpec)

        # create and populate ArchiveInfo instance
        archiveInfo = self.__ArchiveInfo(archiveSpec[_ArchiveSpecOptions.NAME])

        archiveInfo._path = archiveSpec[_ArchiveSpecOptions.PATH]
        archiveInfo._archiverType = archiveSpec[Options.ARCHIVER]
        archiveInfo._destDir = archiveSpec[Options.DEST_DIR]

        if isinstance(archiver, _AbstractIncrementalArchiver):
            archiveInfo._incremental = archiveSpec[Options.INCREMENTAL]

            maxBackupLevel = archiver.getMaxBackupLevel()
            archiveInfo._backupLevel = maxBackupLevel - 1 if maxBackupLevel else None

            archiveInfo._restarting = archiveSpec[Options.RESTARTING]

            archiveInfo._restartAfterLevel = archiveSpec[Options.RESTART_AFTER_LEVEL]

            if not archiver.nextBackupLevel:
                if archiver.restartReason is BackupLevelRestartReasons.RestartCountLimitReached or \
                   archiver.restartReason is BackupLevelRestartReasons.LastFullRestartAgeLimitReached:
                    archiveInfo._nextBackupLevel = 0
                else:
                    archiveInfo._nextBackupLevel = None
            else:
                archiveInfo._nextBackupLevel = archiver.nextBackupLevel
            archiveInfo._restartReason = archiver.restartReason

            archiveInfo._restartLevel = archiver.restartLevel

            if storagePortion.hasVariable(_IncrementalArchiverMixin._STORAGE_RESTART_COUNT):
                archiveInfo._restartCount = int(storagePortion.getValue(
                    _IncrementalArchiverMixin._STORAGE_RESTART_COUNT))

            archiveInfo._fullRestartAfterCount = archiveSpec[Options.FULL_RESTART_AFTER_COUNT]

            if storagePortion.hasVariable(_IncrementalArchiverMixin._STORAGE_LAST_RESTART):
                archiveInfo._lastRestart = _IncrementalArchiverMixin._getRestartDate(
                    _IncrementalArchiverMixin._STORAGE_LAST_RESTART, storagePortion)

            archiveInfo._restartAfterAge = archiveSpec[Options.RESTART_AFTER_AGE]

            if storagePortion.hasVariable(_IncrementalArchiverMixin._STORAGE_LAST_FULL_RESTART):
                archiveInfo._lastFullRestart = _IncrementalArchiverMixin._getRestartDate(
                    _IncrementalArchiverMixin._STORAGE_LAST_FULL_RESTART, storagePortion)

            archiveInfo._fullRestartAfterAge = archiveSpec[Options.FULL_RESTART_AFTER_AGE]

        archiver.detachArchiveSpec()

        return archiveInfo



    def getStoredArchiveInfo(self, archiveName):
        """See: :meth:`.IArchiving.getStoredArchiveInfo()`.

        .. warning:: This method utilizes the :term:`user configuration directory` so the option \
           :attr:`.Options.USER_CONFIG_DIR` has to point to an *existing* directory."""

        # only static methods of archiver classes will be called therefore snapshots directory has to be created by
        # explicit call of the appropriate method
        _ExternalTarArchiver.tryMakeSnapshotsDir(self.__appConfig[Options.USER_CONFIG_DIR],
                                                 self.__interfaceAccessor.getComponentInterface(IComponentUi))

        if archiveName not in self.getStoredArchiveNames():
            return None

        archiveInfo = self.__ArchiveInfo(archiveName)
        storagePortion = self.__storage.createStoragePortion(realm = archiveName)

        archiveInfo._archiverType = self.__appConfig[Options.ARCHIVER]
        archiveInfo._destDir = self.__appConfig[Options.DEST_DIR]

        for archiverClass in reversed(self.__ARCHIVERS):
            if issubclass(archiverClass, _AbstractIncrementalArchiver):
                archiveInfo._backupLevel = archiverClass.getBackupLevelForArchive(
                    archiveName, self.__appConfig[Options.USER_CONFIG_DIR])
                if archiveInfo.backupLevel is not None:
                    break

        archiveInfo._incremental = \
            self.__appConfig[Options.INCREMENTAL] if archiveInfo.backupLevel is not None else None

        if archiveInfo.incremental is not None:
            archiveInfo._restartAfterLevel = self.__appConfig[Options.RESTART_AFTER_LEVEL]

            if storagePortion.hasVariable(_IncrementalArchiverMixin._STORAGE_RESTART_COUNT):
                archiveInfo._restartCount = int(
                    storagePortion.getValue(_IncrementalArchiverMixin._STORAGE_RESTART_COUNT))

            archiveInfo._fullRestartAfterCount = self.__appConfig[Options.FULL_RESTART_AFTER_COUNT]

            if storagePortion.hasVariable(_IncrementalArchiverMixin._STORAGE_LAST_RESTART):
                archiveInfo._lastRestart = _IncrementalArchiverMixin._getRestartDate(
                    _IncrementalArchiverMixin._STORAGE_LAST_RESTART, storagePortion)

            if storagePortion.hasVariable(_IncrementalArchiverMixin._STORAGE_LAST_FULL_RESTART):
                archiveInfo._lastFullRestart = _IncrementalArchiverMixin._getRestartDate(
                    _IncrementalArchiverMixin._STORAGE_LAST_FULL_RESTART, storagePortion)

        return archiveInfo



    @_Utils.uniq
    def getStoredArchiveNames(self):
        """See: :meth:`.IArchiving.getStoredArchiveNames()`.

        .. warning:: This method utilizes the :term:`user configuration directory` so the option \
           :attr:`.Options.USER_CONFIG_DIR` has to point to an *existing* directory."""

        return itertools.chain.from_iterable(
            archiverClass.getStoredArchiveNames(self.__appConfig[Options.USER_CONFIG_DIR], self.__storage)
            for archiverClass in self.__ARCHIVERS)



    def purgeStoredArchiveData(self, archiveName):
        """See: :meth:`.IArchiving.purgeStoredArchiveData()`.

        .. warning:: This method utilizes the :term:`user configuration directory` so the option \
           :attr:`.Options.USER_CONFIG_DIR` has to point to an **existing** directory."""

        purged = False
        for archiverClass in reversed(self.__ARCHIVERS):
            purged = archiverClass.tryPurgeStoredArchiveData(
                archiveName, self.__appConfig[Options.USER_CONFIG_DIR], self.__storage) or purged
        if not purged:
            raise KeyError(str.format("There is no data stored for an archive named \"{}\"."), archiveName)

    # }}} IArchiving implementation



    def __getOrTryCreateArchiveSpec(self, specFile):
        componentUi = self.__interfaceAccessor.getComponentInterface(IComponentUi)
        try:
            return self.__getOrCreateArchiveSpec(specFile)
        except IOError:
            componentUi.showError(str.format("Unable to open archive specification file \"{}\".", specFile))
        except (LookupError, SyntaxError, KeyError, ValueError) as ex:
            componentUi.showError(ex)

        return None



    def __getOrCreateArchiveSpec(self, specFile, silently = False):
        if specFile in self.__archiveSpecs.keys():
            return self.__archiveSpecs[specFile]

        componentUi = self.__interfaceAccessor.getComponentInterface(IComponentUi) if not silently else None

        archiveSpec = _ArchiveSpec(specFile, self.__appConfig, componentUi)
        self.__archiveSpecs[specFile] = archiveSpec
        return archiveSpec



    def __createArchiver(self, archiverType):
        """Creates archiver of type ``archiverType``.

        :param archiverType: Type of the archiver that should be created.
        :type archiverType: :data:`ArchiverTypes`

        :return: An :class:`._AbstractArchiver` instance of the specified concrete type.
        :rtype: :class:`._AbstractArchiver`

        :raise TypeError: If passed ``archiverType`` is not supported."""
 
        try:
            return next((archiver(self.__interfaceAccessor)
                         for archiver in self.__ARCHIVERS
                         if archiver.ARCHIVER_TYPE == archiverType))
        except StopIteration:
            raise TypeError(str.format("Unsupported archiver type: {}.", archiverType))



    class __ArchiveInfo(ArchiveInfo):
        def __init__(self, name):
            super().__init__(name)

# }}} CLASSES
